import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import parent_module
from oarepo_model_builder.validation.utils import ImportSchema

from .defaults import DefaultsModelComponent
from .record_dumper import RecordDumperModelComponent
from .utils import set_default


class EDTFIntervalDumperClassSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    generate = ma.fields.Bool(metadata={"doc": "Generate the dumper class"})
    class_ = ma.fields.Str(
        attribute="class",
        data_key="class",
        metadata={"doc": "Qualified name of the class"},
    )
    base_classes = ma.fields.List(
        ma.fields.Str(),
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "List of base classes"},
    )
    extra_code = ma.fields.Str(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be copied to the bottom of the dumper file"},
    )
    module = ma.fields.String(metadata={"doc": "Class module"})
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "List of python imports"}
    )
    skip = ma.fields.Boolean()


class EDTFIntervalDumperModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [
        DefaultsModelComponent,
        RecordDumperModelComponent,
    ]

    class ModelSchema(ma.Schema):
        edtf_interval_dumper = ma.fields.Nested(
            EDTFIntervalDumperClassSchema,
            attribute="edtf-interval-dumper",
            data_key="edtf-interval-dumper",
            metadata={"doc": "Settings for edtf interval dumper"},
        )

    def before_model_prepare(self, datatype, *, context, **kwargs):
        record_module = parent_module(datatype.definition["record"]["module"])
        prefix = datatype.definition["module"]["prefix"]

        dumper = set_default(datatype, "edtf-interval-dumper", {})
        dumper.setdefault("generate", True)

        dumper_module = dumper.setdefault("module", f"{record_module}.dumpers.edtf")
        ext_class = f"{dumper_module}.{prefix}EDTFIntervalDumperExt"
        dumper.setdefault("class", ext_class)
        dumper.setdefault(
            "base-classes",
            ["oarepo_runtime.records.dumpers.edtf_interval.EDTFIntervalDumperExt"],
        )
        dumper.setdefault("extra-code", "")
        dumper.setdefault("extensions", [])
        dumper.setdefault("imports", [])

        datatype.definition["record-dumper"]["extensions"].append(
            "{{" + ext_class + "}}()"
        )



import marshmallow as ma
from marshmallow import fields

from oarepo_model_builder.datatypes import DataTypeComponent, ObjectDataType
from oarepo_model_builder.validation.extensibility import ExtensibleSchema
from oarepo_model_builder.validation.utils import StrictSchema

from .marshmallow import ObjectMarshmallowSchema, PropertyMarshmallowSchema


class PropertyUISchema(StrictSchema):
    marshmallow = fields.Nested(
        PropertyMarshmallowSchema, metadata={"doc": "UI marshmallow"}
    )


class RegularUIComponent(DataTypeComponent):
    class ModelSchema(ma.Schema):
        ui = ma.fields.Nested(
            ExtensibleSchema("ui.property", PropertyUISchema),
            required=False,
        )


class ObjectUIExtraSchema(ma.Schema):
    marshmallow = fields.Nested(ObjectMarshmallowSchema)


class ObjectUISchema(ObjectUIExtraSchema, PropertyUISchema):
    pass


class ObjectUIComponent(RegularUIComponent):
    eligible_datatypes = [ObjectDataType]

    class ModelSchema(ma.Schema):
        ui = ma.fields.Nested(ExtensibleSchema("ui.object", ObjectUISchema))



import logging
from collections import namedtuple

default_logger = logging.getLogger("oarepo_model_builder")


# verbosity 1 is the most serious
# verbosity 9 is the least serious


class Log:
    INFO = 0
    ERROR = logging.INFO - logging.ERROR

    LogStackEntry = namedtuple("LogStackEntry", "verbosity,indent,logger")

    def __init__(self):
        self.stack = [Log.LogStackEntry(verbosity=0, indent=0, logger=default_logger)]

    def enter(self, verbosity, fmt, *args, logger=None, **kwargs):
        if not logger:
            logger = self.stack[-1].logger
        level = self.stack[-1].indent

        if logger.isEnabledFor(logging.INFO - verbosity):
            self(verbosity, fmt, *args, logger=logger, **kwargs)
            level += 1

        self.stack.append(
            Log.LogStackEntry(verbosity=verbosity, indent=level, logger=logger)
        )

    def leave(self, fmt=None, *args, **kwargs):
        top = self.stack.pop()
        if fmt:
            self(top.verbosity, fmt, *args, logger=top.logger, **kwargs)

    def __call__(self, verbosity, fmt, *args, logger=None, **kwargs):
        if not logger:
            logger = self.stack[-1].logger
        indent = "  " * (self.stack[-1].indent - 1)
        logger.log(logging.INFO - verbosity, indent + fmt, *args, **kwargs)


log = Log()



import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import parent_module
from oarepo_model_builder.validation.utils import ImportSchema

from .defaults import DefaultsModelComponent
from .record import RecordModelComponent
from .utils import set_default


class RecordMetadataClassSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    alias = ma.fields.Str(
        metadata={"doc": "Alias under which the metadata is registered in setup.cfg"}
    )
    generate = ma.fields.Bool(metadata={"doc": "True to generate the metadata class"})
    module = ma.fields.String(
        metadata={"doc": "Module where the metadata class resides"}
    )
    class_ = ma.fields.Str(
        attribute="class",
        data_key="class",
        metadata={"doc": "Qualified name of the class"},
    )
    base_classes = ma.fields.List(
        ma.fields.Str(),
        attribute="base-classes",
        data_key="base-classes",
        metadata={"doc": "A list of base classes"},
    )
    extra_code = ma.fields.Str(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be put below the record"},
    )
    table = ma.fields.Str(
        attribute="table",
        data_key="table",
        metadata={"doc": "Name of the database table"},
    )
    alembic = ma.fields.Str(
        metadata={
            "doc": (
                "module where alembic files are stored. Used only for records profile, "
                "ignored in other profiles as alembic is always stored inside a record"
            )
        }
    )
    use_versioning = ma.fields.Boolean(
        attribute="use-versioning", data_key="use-versioning", load_default=True
    )
    imports = ma.fields.List(
        ma.fields.Nested(ImportSchema), metadata={"doc": "List of python imports"}
    )
    skip = ma.fields.Boolean()


class RecordMetadataModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent, RecordModelComponent]

    class ModelSchema(ma.Schema):
        record_metadata = ma.fields.Nested(
            RecordMetadataClassSchema,
            attribute="record-metadata",
            data_key="record-metadata",
            metadata={"doc": "Record metadata settings"},
        )

    def before_model_prepare(self, datatype, context, **kwargs):
        profile = context["profile"]
        records_module = parent_module(datatype.definition["record"]["module"])
        prefix = datatype.definition["module"]["prefix"]
        alias = datatype.definition["module"]["alias"]

        metadata = set_default(datatype, "record-metadata", {})
        metadata.setdefault("generate", True)
        metadata_module = metadata.setdefault("module", f"{records_module}.models")
        metadata.setdefault("class", f"{metadata_module}.{prefix}Metadata")
        metadata.setdefault(
            "base-classes",
            [
                "invenio_db.db{db.Model}",
                "invenio_records.models.RecordMetadataBase",
            ],
        )
        metadata.setdefault("extra-code", "")
        metadata.setdefault(
            "imports",
            [],
        )
        metadata.setdefault(
            "table", f"{datatype.definition['module']['prefix-snake']}_metadata"
        )
        metadata.setdefault("alias", alias)
        metadata.setdefault("use-versioning", True)

        if profile == "record":
            # only add alembic for records profile, do not add it for other profiles.
            # we can not split alembic as both are installed at the same time and
            # alembic autodetection places changes just to the first one
            alembic_module = datatype.definition["module"]["qualified"] + ".alembic"
            metadata.setdefault("alembic", alembic_module)



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordDumperBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_dumper"
    section = "record-dumper"
    template = "record-dumper"



import re

from marshmallow import fields
from marshmallow.exceptions import ValidationError

from .datatypes import DataType


def validate_regex(value):
    if not value:
        return
    try:
        re.compile(value)
    except Exception as e:
        raise ValidationError(
            f"Regex '{value}' is not valid. Reported error: '{str(e)}'"
        )


class StringDataType(DataType):
    ui = {
        "marshmallow": {
            "field-class": "marshmallow.fields{ma_fields.String}",
        }
    }
    marshmallow = {
        "field-class": "marshmallow.fields{ma_fields.String}",
    }
    json_schema = {"type": "string"}

    class ModelSchema(DataType.ModelSchema):
        minLength = fields.Integer(required=False)
        maxLength = fields.Integer(required=False)
        regex = fields.String(required=False, validate=validate_regex)


class FulltextDataType(StringDataType):
    model_type = "fulltext"
    mapping = {"type": "text"}


class KeywordDataType(StringDataType):
    model_type = "keyword"
    mapping = {"type": "keyword", "ignore_above": 1024}
    facets = {
        "facet-class": "invenio_records_resources.services.records.facets.TermsFacet",
    }


class UUIDDataType(StringDataType):
    model_type = "uuid"
    mapping = {"type": "keyword"}


class FulltextKeywordDataType(StringDataType):
    model_type = "fulltext+keyword"
    mapping = {
        "type": "text",
        "fields": {"keyword": {"type": "keyword", "ignore_above": 256}},
    }
    facets = {"facet": False}


class URLDataType(StringDataType):
    model_type = "url"
    mapping = {"type": "keyword", "ignore_above": 1024}
    facets = {
        "facet-class": "invenio_records_resources.services.records.facets.TermsFacet",
    }



TEMPLATES = {
    "record": "templates/record.py.jinja2",
    "record-metadata": "templates/record_metadata.py.jinja2",
    "permissions": "templates/permissions.py.jinja2",
    "record-search-options": "templates/invenio_record_search_options.py.jinja2",
    "service-config": "templates/service_config.py.jinja2",
    "service": "templates/service.py.jinja2",
    "marshmallow": "templates/marshmallow.py.jinja2",
    "record-dumper": "templates/record_dumper.py.jinja2",
    "ui-serializer": "templates/ui_serializer.py.jinja2",
    "resource-config": "templates/resource_config.py.jinja2",
    "resource": "templates/resource.py.jinja2",
    "api-views": "templates/api_views.py.jinja2",
    "app-views": "templates/app_views.py.jinja2",
    "version": "templates/version.py.jinja2",
    "config": "templates/config.py.jinja2",
    "ext": "templates/ext.py.jinja2",
    "proxies": "templates/proxies.py.jinja2",
    "record-facets": "templates/invenio_record_facets.py.jinja2",
    "pid-provider": "templates/pid_provider.py.jinja2",
    # utils and included
    "ext-resource": "templates/ext_resource.py.jinja2",
    "edtf-interval-record-dumper": "templates/edtf_interval_record_dumper.py.jinja2",
    "record-item": "templates/record_item.py.jinja2",
    "record-list": "templates/record_list.py.jinja2",
}



from oarepo_model_builder.utils.jinja import split_package_base_name

from ..builders import OutputBuilder
from ..outputs.cfg import CFGOutput


class InvenioExtSetupCfgBuilder(OutputBuilder):
    TYPE = "invenio_ext_setup_cfg"

    def finish(self):
        super().finish()

        output: CFGOutput = self.builder.get_output("cfg", "setup.cfg")

        package_name, base_name = split_package_base_name(
            self.current_model.definition["ext"]["class"]
        )

        output.add_entry_point(
            "invenio_base.api_apps",
            self.current_model.definition["ext"]["alias"],
            f"{package_name}:{base_name}",
        )

        # need to add ext to apps because cli depends on it
        output.add_entry_point(
            "invenio_base.apps",
            self.current_model.definition["ext"]["alias"],
            f"{package_name}:{base_name}",
        )



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioAPIViewsBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_api_views"
    section = "api-blueprint"
    template = "api-views"

    def finish(self, **extra_kwargs):
        ext = self.current_model.section_ext_resource.config
        super().finish(ext=ext, **extra_kwargs)



from typing import Iterator

from oarepo_model_builder.invenio.invenio_base import InvenioBaseClassPythonBuilder


class EDTFIntervalDumperBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "edtf_interval_dumper"
    section = "edtf-interval-dumper"
    template = "edtf-interval-record-dumper"

    def finish(self, **extra_kwargs):
        paths = list(self.get_paths(self.current_model.children))
        extra_kwargs["paths"] = sorted(set(paths))
        super().finish(**extra_kwargs)

    def get_paths(self, parent_node) -> Iterator[str]:
        children = parent_node
        for c in children:
            node = children[c]
            if node.model_type == "edtf-interval":
                yield self.to_path(node.path)
            elif node.children != {}:
                yield from self.get_paths(node.children)
            elif hasattr(node, "item"):
                yield from self.get_paths({"item": node.item})

    def to_path(self, path):
        components = path.split(".")
        return "/".join(components)



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordResourceBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_resource"
    section = "resource"
    template = "resource"



from pathlib import Path
from typing import List, Union

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.profiles import Profile
from oarepo_model_builder.schema import ModelSchema
from oarepo_model_builder.utils.dict import dict_get


class RecordProfile(Profile):
    default_model_path = ("record",)

    def build(
        self,
        model: ModelSchema,
        profile: str,
        model_path: List[str],
        output_directory: Union[str, Path],
        builder: ModelBuilder,
        **kwargs,
    ):
        current_model = dict_get(model.schema, model_path)

        # record has type "model" if not explicitly stated otherwise
        if "type" not in current_model:
            current_model["type"] = "model"

        return super().build(
            model, profile, model_path, output_directory, builder, **kwargs
        )



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordServiceBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_service"
    section = "service"
    template = "service"



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioVersionBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_version"
    section = "version"
    template = "version"
    generate = True

    def _get_output_module(self):
        return f"{self.current_model.definition['module']['qualified']}.version"



from configupdater import ConfigUpdater, Option

from oarepo_model_builder.outputs import OutputBase


class CFGOutput(OutputBase):
    TYPE = "cfg"

    # high-level API

    def add_entry_point(self, group, name, value):
        """
        Adds an entry point if it is not already present
        """
        line = f"{name} = {value}"
        grp = self.create_multiline_value(
            "options.entry_points", group, initial_value=line
        )
        for e in grp.as_list():
            e = [x.strip() for x in e.split("=", maxsplit=1)]
            if len(e) == 2 and e[0] == name and e[1] == value:
                break
        else:
            grp.append(line)

    def add_dependency(
        self, package, version, group="options", section="install_requires", extras=None
    ):
        if extras:
            line = f'{package}[{", ".join(extras)}]{version}'
        else:
            line = f"{package}{version}"
        grp = self.create_multiline_value(group, section, line)
        for e in grp.as_list():
            if line == e.strip():
                break
        else:
            grp.append(line)

    # low-level API

    def begin(self):
        try:
            with self.builder.filesystem.open(self.path) as f:
                self.original_data = f.read()
                self.cfg = ConfigUpdater()
                self.cfg.read_string(self.original_data)
                self.parsed = str(self.cfg)
        except FileNotFoundError:
            self.original_data = None
            self.cfg = ConfigUpdater()
            self.parsed = None

    @property
    def created(self):
        return self.original_data is None

    def get(self, section, key):
        try:
            tbl = self.cfg[section]
            return tbl[key]
        except KeyError:
            return None

    def get_section(self, section, create=False):
        if create and section not in self.cfg:
            self.cfg.add_section(section)
            tbl = self.cfg[section]
            tbl.add_before.space(2)
            return tbl
        return self.cfg[section]

    def set(self, section, key, value):
        tbl = self.get_section(section, create=True)
        tbl[key] = value

    def setdefault(self, section, key, value):
        tbl = self.get_section(section, create=True)
        return tbl.setdefault(key, value)

    def create_multiline_value(self, section, name, initial_value="") -> Option:
        tbl = self.get_section(section, create=True)
        if name in tbl:
            grp = tbl[name]
        else:
            grp = Option(name, value=initial_value)
            tbl.add_option(grp)
        return grp

    def finish(self):
        out = str(self.cfg)
        if out != self.parsed:
            with self.builder.filesystem.open(self.path, "w") as f:
                f.write(out)



from ..builders import OutputBuilder
from ..outputs.toml import TOMLOutput


class PyprojectTOMLBuilder(OutputBuilder):
    TYPE = "pyproject_toml"

    def finish(self):
        super().finish()
        output: TOMLOutput = self.builder.get_output("toml", "pyproject.toml")
        output.setdefault(
            "build_system", "requires", ["setuptools", "wheel", "babel>2.8"]
        )
        output.setdefault("build_system", "build-backend", "setuptools.build_meta")
